"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const path_1 = require("path");
const source_map_support_1 = require("source-map-support");
const chalk = require("chalk");
const util_1 = require("util");
const LoggerHelper_1 = require("./LoggerHelper");
class Logger {
    constructor(settings) {
        var _a, _b, _c, _d, _e, _f, _g, _h, _j, _k;
        this._logLevels = {
            0: "silly",
            1: "trace",
            2: "debug",
            3: "info",
            4: "warn",
            5: "error",
            6: "fatal",
        };
        this._ignoreStackLevels = 3;
        this._attachedTransports = [];
        this._minLevelToStdErr = 4;
        this.settings = {
            instanceId: settings === null || settings === void 0 ? void 0 : settings.instanceId,
            name: (_a = settings === null || settings === void 0 ? void 0 : settings.name) !== null && _a !== void 0 ? _a : "",
            minLevel: (_b = settings === null || settings === void 0 ? void 0 : settings.minLevel) !== null && _b !== void 0 ? _b : 0,
            exposeStack: (_c = settings === null || settings === void 0 ? void 0 : settings.exposeStack) !== null && _c !== void 0 ? _c : false,
            suppressLogging: (_d = settings === null || settings === void 0 ? void 0 : settings.suppressLogging) !== null && _d !== void 0 ? _d : false,
            overwriteConsole: (_e = settings === null || settings === void 0 ? void 0 : settings.overwriteConsole) !== null && _e !== void 0 ? _e : false,
            logAsJson: (_f = settings === null || settings === void 0 ? void 0 : settings.logAsJson) !== null && _f !== void 0 ? _f : false,
            logLevelsColors: (_g = settings === null || settings === void 0 ? void 0 : settings.logLevelsColors) !== null && _g !== void 0 ? _g : {
                0: "#B0B0B0",
                1: "#FFFFFF",
                2: "#63C462",
                3: "#2020C0",
                4: "#CE8743",
                5: "#CD444C",
                6: "#FF0000",
            },
            jsonHighlightColors: (_h = settings === null || settings === void 0 ? void 0 : settings.jsonHighlightColors) !== null && _h !== void 0 ? _h : {
                number: "#FF6188",
                key: "#A9DC76",
                string: "#FFD866",
                boolean: "#FC9867",
                null: "#AB9DF2",
            },
            stdOut: (_j = settings === null || settings === void 0 ? void 0 : settings.stdOut) !== null && _j !== void 0 ? _j : process.stdout,
            stdErr: (_k = settings === null || settings === void 0 ? void 0 : settings.stdErr) !== null && _k !== void 0 ? _k : process.stderr,
        };
        LoggerHelper_1.LoggerHelper.errorToJsonHelper();
        if (this.settings.overwriteConsole) {
            LoggerHelper_1.LoggerHelper.overwriteConsole(this, this._handleLog);
        }
    }
    attachTransport(logger, minLevel = 0) {
        this._attachedTransports.push({
            minLevel,
            logger,
        });
    }
    silly(...args) {
        return this._handleLog.apply(this, [0, args]);
    }
    trace(...args) {
        return this._handleLog.apply(this, [1, args, true]);
    }
    debug(...args) {
        return this._handleLog.apply(this, [2, args]);
    }
    info(...args) {
        return this._handleLog.apply(this, [3, args]);
    }
    warn(...args) {
        return this._handleLog.apply(this, [4, args]);
    }
    error(...args) {
        return this._handleLog.apply(this, [5, args]);
    }
    /**
     * Returns the average of two numbers.
     *
     * @param x - The first input number
     * @returns LogObject containing all the relevant information about this log
     *
     */
    fatal(...args) {
        return this._handleLog.apply(this, [6, args]);
    }
    /** @internal */
    _handleLog(logLevel, logArguments, doExposeStack = this.settings.exposeStack) {
        const logObject = this._buildLogObject(logLevel, logArguments, doExposeStack);
        if (!this.settings.suppressLogging && logLevel >= this.settings.minLevel) {
            if (!this.settings.logAsJson) {
                this._printPrettyLog(logObject);
            }
            else {
                this._printJsonLog(logObject);
            }
        }
        this._attachedTransports.forEach((transport) => {
            if (logLevel >= transport.minLevel) {
                transport.logger[this._logLevels[logLevel]](logObject);
            }
        });
        return logObject;
    }
    /** @internal */
    _buildLogObject(logLevel, logArguments, doExposeStack = true) {
        var _a;
        const callSites = LoggerHelper_1.LoggerHelper.getCallSites();
        const relevantCallSites = callSites.splice(this._ignoreStackLevels);
        const stackFrame = source_map_support_1.wrapCallSite(relevantCallSites[0]);
        const stackFrameObject = LoggerHelper_1.LoggerHelper.toStackFrameObject(stackFrame);
        const logObject = {
            loggerName: (_a = this.settings.name) !== null && _a !== void 0 ? _a : "",
            date: new Date(),
            logLevel: logLevel,
            logLevelName: this._logLevels[logLevel],
            filePath: stackFrameObject.filePath,
            fullFilePath: stackFrameObject.fullFilePath,
            fileName: stackFrameObject.fileName,
            lineNumber: stackFrameObject.lineNumber,
            columnNumber: stackFrameObject.columnNumber,
            isConstructor: stackFrameObject.isConstructor,
            functionName: stackFrameObject.functionName,
            typeName: stackFrameObject.typeName,
            methodName: stackFrameObject.methodName,
            argumentsArray: [],
        };
        logArguments.forEach((arg) => {
            var _a;
            if (typeof arg === "object" && !util_1.types.isNativeError(arg)) {
                logObject.argumentsArray.push(JSON.parse(JSON.stringify(arg)));
            }
            else if (typeof arg === "object" && util_1.types.isNativeError(arg)) {
                const errorStack = LoggerHelper_1.LoggerHelper.getCallSites(arg);
                const errorObject = JSON.parse(JSON.stringify(arg));
                errorObject.name = (_a = errorObject.name) !== null && _a !== void 0 ? _a : "Error";
                errorObject.isError = true;
                errorObject.stack = this._toStackObjectArray(errorStack);
                logObject.argumentsArray.push(errorObject);
            }
            else {
                logObject.argumentsArray.push(util_1.format(arg));
            }
        });
        if (doExposeStack) {
            logObject.stack = this._toStackObjectArray(relevantCallSites);
        }
        return logObject;
    }
    /** @internal */
    _toStackObjectArray(jsStack) {
        const prettyStack = Object.values(jsStack).reduce((iPrettyStack, stackFrame) => {
            iPrettyStack.push(LoggerHelper_1.LoggerHelper.toStackFrameObject(source_map_support_1.wrapCallSite(stackFrame)));
            return iPrettyStack;
        }, []);
        return prettyStack;
    }
    /** @internal */
    _printPrettyLog(logObject) {
        const std = logObject.logLevel < this._minLevelToStdErr
            ? this.settings.stdOut
            : this.settings.stdErr;
        const nowStr = logObject.date
            .toISOString()
            .replace("T", " ")
            .replace("Z", "");
        const hexColor = this.settings.logLevelsColors[logObject.logLevel];
        std.write(chalk `{grey ${nowStr}}\t`);
        std.write(chalk.hex(hexColor).bold(` ${logObject.logLevelName.toUpperCase()}\t`));
        const functionName = logObject.isConstructor
            ? ` ${logObject.typeName}.constructor`
            : logObject.methodName != null
                ? ` ${logObject.typeName}.${logObject.methodName}`
                : logObject.functionName != null
                    ? ` ${logObject.functionName}`
                    : "";
        const optionalInstanceId = this.settings.instanceId != null ? `@${this.settings.instanceId}` : "";
        std.write(chalk.gray(`[${logObject.loggerName}${optionalInstanceId} ${logObject.filePath}:${logObject.lineNumber}${functionName}]\t`));
        logObject.argumentsArray.forEach((argument) => {
            const errorArgument = argument;
            if (typeof argument === "object" && !errorArgument.isError) {
                std.write("\n" +
                    LoggerHelper_1.LoggerHelper.colorizeJson(argument, chalk, this.settings.jsonHighlightColors) +
                    " ");
            }
            else if (typeof argument === "object" && errorArgument.isError) {
                std.write(util_1.format(chalk `\n{whiteBright.bgRed.bold ${errorArgument.name}}{grey :} ${util_1.format(errorArgument.message)}\n`));
                this._printPrettyStack(std, errorArgument.stack);
            }
            else {
                std.write(util_1.format(argument + " "));
            }
        });
        std.write("\n");
        if (logObject.stack != null) {
            std.write(chalk `{underline.bold log stack:\n}`);
            this._printPrettyStack(std, logObject.stack);
        }
    }
    /** @internal */
    _printPrettyStack(std, stackObjectArray) {
        std.write("\n");
        Object.values(stackObjectArray).forEach((stackObject) => {
            var _a;
            std.write(chalk `    {grey •} {yellowBright ${stackObject.fileName}}{grey :}{yellow ${stackObject.lineNumber}} {white ${(_a = stackObject.functionName) !== null && _a !== void 0 ? _a : "<anonumous>"}}`);
            std.write("\n    ");
            std.write(path_1.normalize(chalk `{grey ${stackObject.fullFilePath}:${stackObject.lineNumber}:${stackObject.columnNumber}}`));
            std.write("\n\n");
        });
    }
    /** @internal */
    _printJsonLog(logObject) {
        const std = logObject.logLevel < this._minLevelToStdErr
            ? this.settings.stdOut
            : this.settings.stdErr;
        std.write(LoggerHelper_1.LoggerHelper.colorizeJson(JSON.stringify(logObject), chalk, this.settings.jsonHighlightColors) + "\n");
    }
}
exports.Logger = Logger;
//# sourceMappingURL=index.js.map